#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


# shape class definition for symbol editor.
# base class for all shapes: rectangle, circle, line
import itertools
import math
from pathlib import Path
from typing import Tuple, Union

import numpy as np
from PySide6.QtCore import (
    QLineF,
    QPoint,
    QPointF,
    QRect,
    QRectF,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QFont,
    QFontDatabase,
    QFontMetrics,
    QImage,
    QPainter,
    QPainterPath,
    QPen,
    QPixmap,
    QPolygonF,
    QTextOption,
    QTransform,
)
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsSceneHoverEvent,
    QGraphicsSceneMouseEvent,
    QStyle,
)

import revedaEditor.backend.dataDefinitions as ddef
from revedaEditor.backend.pdkLoader import importPDKModule


class textureCache:
    _file_content_cache = {}
    _pixmap_cache = {}

    @classmethod
    def readFileContent(cls, filePath):
        if filePath not in cls._file_content_cache:
            with open(filePath, "r") as file:
                cls._file_content_cache[filePath] = file.read()
        return cls._file_content_cache[filePath]

    @classmethod
    def createImage(cls, filePath: Path, color: QColor, scale: int = 1) -> QImage:
        content = cls.readFileContent(str(filePath))

        # Use numpy's loadtxt for faster parsing of text data
        data = np.loadtxt(content.splitlines(), dtype=np.uint8)

        # Scale up the pattern by repeating each pixel
        data_scaled = np.repeat(np.repeat(data, scale, axis=0), scale, axis=1)

        height, width = data_scaled.shape

        # Create QImage with Format_ARGB32 (not premultiplied)
        image = QImage(width, height, QImage.Format.Format_ARGB32)
        # Fill with transparent pixels first
        image.fill(Qt.transparent)

        # Create painter to draw on the image
        painter = QPainter(image)
        if not painter.isActive():  # Add this check
            return image
        painter.setPen(Qt.NoPen)
        # painter.setBrush(QBrush(color))
        # Create semi-transparent color (50% opacity)
        transparent_color = QColor(color)
        transparent_color.setAlpha(230)  # 220 is 90% opacity (range is 0-255)
        painter.setBrush(QBrush(transparent_color))

        # Draw solid rectangles for each pixel that should be colored
        for i in range(height):
            for j in range(width):
                if data_scaled[i, j] == 1:  # Draw colored pixel
                    painter.drawRect(j, i, 1, 1)

        painter.end()
        return image

    @classmethod
    def getCachedPixmap(cls, texturePath, color):
        cache_key = (str(texturePath), color.name())
        if cache_key not in cls._pixmap_cache:
            image = cls.createImage(texturePath, color, 1)
            pixmap = QPixmap.fromImage(image)
            cls._pixmap_cache[cache_key] = pixmap
        return cls._pixmap_cache[cache_key]

    @classmethod
    def clearCaches(cls):
        cls._file_content_cache.clear()
        cls._pixmap_cache.clear()


class layoutShape(QGraphicsItem):
    # Class-level color cache
    _color_cache = {}
    # Class-level pen/brush lookup table
    _pen_brush_cache = {}

    def __init__(self) -> None:
        super().__init__()

        # Batch flag operations for better performance
        flags = (
                QGraphicsItem.ItemIsSelectable
                | QGraphicsItem.ItemSendsGeometryChanges
                | QGraphicsItem.ItemIsFocusable
                | QGraphicsItem.ItemUsesExtendedStyleOption
        )
        self.setFlags(flags)

        # Single method calls
        self.setAcceptHoverEvents(True)
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

        # Direct attribute initialization (faster than individual assignments)
        self._pen = self._brush = None
        self._angle = 0
        self._stretch = False
        self._offset = QPoint(0, 0)
        self._flipTuple = (1, 1)
        self._transformedBrush = self._lastScale = None

    @classmethod
    def _get_cached_color(cls, color_name):
        if color_name not in cls._color_cache:
            cls._color_cache[color_name] = QColor(color_name)
        return cls._color_cache[color_name]

    def __repr__(self) -> str:
        return "layoutShape()"

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedHasChanged and self.scene():
            # Direct z-value calculation without conditional
            self.setZValue(self.zValue() + (20 * value - 10))
        return super().itemChange(change, value)

    def _definePensBrushes(self, layer):
        # Create cache key from layer properties
        cache_key = (
            layer.name,
            layer.purpose,
            layer.pcolor.name(),
            layer.pwidth,
            layer.pstyle,
            layer.bcolor.name(),
            layer.btexture,
        )

        if cache_key not in self._pen_brush_cache:
            # Import PDK module when needed to avoid circular imports
            laylyr = importPDKModule("layoutLayers")
            # Create objects only once per unique layer configuration
            texturePath = Path(laylyr.__file__).parent.joinpath(layer.btexture)
            _pixmap = textureCache.getCachedPixmap(texturePath, layer.bcolor)

            yellow = self._get_cached_color("yellow")
            red = self._get_cached_color("red")

            self._pen_brush_cache[cache_key] = {
                "pen": QPen(layer.pcolor, layer.pwidth, layer.pstyle),
                "brush": QBrush(layer.bcolor, _pixmap),
                "selectedPen": QPen(yellow, layer.pwidth, Qt.DashLine),
                "selectedBrush": QBrush(yellow, _pixmap),
                "stretchPen": QPen(red, layer.pwidth, Qt.SolidLine),
                "stretchBrush": QBrush(red, _pixmap),
            }

        # Assign from cache
        cached = self._pen_brush_cache[cache_key]
        self._pen = cached["pen"]
        self._brush = cached["brush"]
        self._selectedPen = cached["selectedPen"]
        self._selectedPen.setCosmetic(True)
        self._selectedBrush = cached["selectedBrush"]
        self._stretchPen = cached["stretchPen"]
        self._stretchPen.setCosmetic(True)
        self._stretchBrush = cached["stretchBrush"]

    def _updateTransformedBrush(self, brush: QBrush, scale: float):
        """Update transformed brush only when needed"""
        rounded_scale = max(round(scale, 2), 0.01)  # Prevent division by zero

        if self._transformedBrush is None or self._lastScale != rounded_scale:
            if self._transformedBrush is None:
                self._transformedBrush = QBrush(brush)
            else:
                self._transformedBrush = brush

            transform = QTransform().scale(1 / rounded_scale, 1 / rounded_scale)
            self._transformedBrush.setTransform(transform)
            self._lastScale = rounded_scale

    @property
    def pen(self):
        return self._pen

    @pen.setter
    def pen(self, value: QPen):
        if isinstance(value, QPen):
            self._pen = value

    @property
    def brush(self):
        return self._brush

    @brush.setter
    def brush(self, value: QBrush):
        if isinstance(value, QBrush):
            self._brush = value

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.prepareGeometryChange()
        self.setRotation(value)

    @property
    def stretch(self):
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value

    @property
    def view(self):
        if self.scene():
            return self.scene().views()[0]
        else:
            return None

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value: Union[QPoint | QPointF]):
        self._offset = value

    # def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
    #     super().mousePressEvent(event)
    #     if self.scene() and self.scene().editModes.moveItem:
    #         self.setFlag(QGraphicsItem.ItemIsMovable, True)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setSelected(True)
        if self.scene() and self.scene().editModes.moveItem:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)

        super().mousePressEvent(event)

    def sceneEvent(self, event):
        """
        Do not propagate event if shape needs to keep still.
        """
        if self.scene() and (
                self.scene().editModes.changeOrigin or self.scene().drawMode
        ):
            return False
        else:
            super().sceneEvent(event)
            return True

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self.setCursor(Qt.ArrowCursor)
        self.setOpacity(0.75)
        self.setFocus()
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        self.setCursor(Qt.CrossCursor)
        self.setOpacity(1)
        self.clearFocus()

    def contextMenuEvent(self, event):
        self.setSelected(True)
        self.scene().itemContextMenu.exec_(event.screenPos())

    @property
    def flipTuple(self):
        return self._flipTuple

    @flipTuple.setter
    def flipTuple(self, flipState: Tuple[int, int]):
        # Early exit if no change
        if self._flipTuple == flipState:
            return

        # Cache center calculation
        if not hasattr(self, "_cached_center"):
            self._cached_center = self.boundingRect().center()
        center = self._cached_center

        # Direct transform creation is faster than modifying existing
        cx, cy = center.x(), center.y()
        sx, sy = flipState

        transform = QTransform(sx, 0, 0, sy, cx - sx * cx, cy - sy * cy)
        self.setTransform(transform)
        self._flipTuple = flipState

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = value
        self._definePensBrushes(self._layer)


class layoutRect(layoutShape):
    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(
            self,
            start: QPoint,
            end: QPoint,
            layer: ddef.layLayer,
    ):
        super().__init__()
        self._rect = QRectF(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._layer = layer
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)
            self.setFlag(QGraphicsItem.ItemIsFocusable, False)
        self._stretch = False
        self._stretchSide = None
        self._stretchPen = QPen(QColor("red"), self._layer.pwidth, Qt.SolidLine)
        self._definePensBrushes(self._layer)
        self.setZValue(self._layer.z)
        self._stretchSidesMap = {
            self.sides[0]: (lambda r: r.topLeft(), lambda r: r.bottomLeft()),
            self.sides[1]: (lambda r: r.topRight(), lambda r: r.bottomRight()),
            self.sides[2]: (lambda r: r.topLeft(), lambda r: r.topRight()),
            self.sides[3]: (lambda r: r.bottomLeft(), lambda r: r.bottomRight()),
        }

    def __repr__(self):
        return f"layoutRect({self._start}, {self._end}, {self._layer})"

    def paint(self, painter, option, widget):
        rect = self._rect

        # Get scale once and cache it
        scale = self.scene().views()[0].transform().m11()

        if self.isSelected():
            painter.setPen(self._selectedPen)
            self._updateTransformedBrush(self._selectedBrush, scale)

            if self.stretch:
                painter.setPen(self._stretchPen)
                # Get the line endpoints from the mapping
                if self._stretchSide in self._stretchSidesMap:
                    start_func, end_func = self._stretchSidesMap[self._stretchSide]
                    painter.drawLine(start_func(rect), end_func(rect))
        else:
            painter.setPen(self._pen)
            self._updateTransformedBrush(self._brush, scale)
        painter.setBrush(self._transformedBrush)
        painter.drawRect(rect)

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self.end).normalized()
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self.start, end).normalized()
        self._end = end

    @property
    def centre(self):
        return QPoint(
            int(self._rect.x() + self._rect.width() / 2),
            int(self._rect.y() + self._rect.height() / 2),
        )

    @property
    def height(self):
        return self.rect.height()

    @height.setter
    def height(self, height: int):
        self.prepareGeometryChange()
        self._rect.setHeight(height)

    @property
    def width(self):
        return self.rect.width()

    @width.setter
    def width(self, width):
        self.prepareGeometryChange()
        self.rect.setWidth(width)

    @property
    def left(self):
        return self.rect.left()

    @left.setter
    def left(self, left: int):
        self.rect.setLeft(left)

    @property
    def right(self):
        return self.rect.right()

    @right.setter
    def right(self, right: int):
        self.prepareGeometryChange()
        self.rect.setRight(right)

    @property
    def top(self):
        return self.rect.top()

    @top.setter
    def top(self, top: int):
        self.prepareGeometryChange()
        self.rect.setTop(top)

    @property
    def bottom(self):
        return self.rect.bottom()

    @bottom.setter
    def bottom(self, bottom: int):
        self.prepareGeometryChange()
        self.rect.setBottom(bottom)

    @property
    def origin(self):
        return self.rect.bottomLeft()

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, layer: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = layer

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)

        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                if eventPos.x() == self._rect.left():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[0]
                elif eventPos.x() == self._rect.right():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[1]
                elif eventPos.y() == self._rect.top():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[2]
                elif eventPos.y() == self._rect.bottom():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[3]
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self.stretch:
            self.prepareGeometryChange()
            if self.stretchSide == layoutRect.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == layoutRect.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self._pen.width() / 2))
            elif self.stretchSide == layoutRect.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == layoutRect.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self._pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class layoutInstance(layoutShape):
    def __init__(self, shapes: list[layoutShape]):
        super().__init__()

        # Direct attribute initialization
        self._shapes = shapes
        self._draft = False
        self._libraryName = self._cellName = self._viewName = self._instanceName = ""
        self._counter = 0

        # Cache pen creation
        self._selectedPen = self._get_cached_color("yellow")
        pen = QPen(self._selectedPen, 4, Qt.DashLine)
        pen.setCosmetic(True)
        self._selectedPen = pen

        # Batch flag operations
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)

        # Defer expensive operations
        self._start = None
        self._shapes_set = False

        # Set shapes only if not empty
        if shapes:
            self.setShapes()

    def setShapes(self):
        if self._shapes_set or not self._shapes:
            return

        # Batch process all shapes
        for item in self._shapes:
            item.setFlags(QGraphicsItem.ItemStacksBehindParent)
            item.setFlag(QGraphicsItem.ItemIsSelectable, False)
            item.setParentItem(self)

        self._shapes_set = True

    def removeShapes(self):
        self.prepareGeometryChange()
        for item in self._shapes:
            item.setParentItem(None)
            del item
        self._shapes = list()

    def __repr__(self):
        return f"{self.__class__.__name__}({self._libraryName}, {self._cellName}, {self._viewName}, {self._instanceName})"

    def boundingRect(self):
        return self.childrenBoundingRect().normalized().adjusted(-2, -2, 2, 2)

    def paint(self, painter, option, widget):
        painter.setRenderHint(QPainter.NonCosmeticBrushPatterns)
        if option.state & QStyle.State_Selected:
            # if self in self.scene().selectedItemsSet:
            painter.setPen(self._selectedPen)
            painter.drawRect(self.childrenBoundingRect())

    def sceneEvent(self, event):
        """
        Do not propagate event if shape needs to keep still.
        """
        if not (
                self.scene().selectModes.selectInstance
                or self.scene().selectModes.selectAll
        ):
            return False
        else:
            super().sceneEvent(event)
            return True

    @property
    def libraryName(self):
        return self._libraryName

    @libraryName.setter
    def libraryName(self, value):
        self._libraryName = value

    @property
    def cellName(self):
        return self._cellName

    @cellName.setter
    def cellName(self, value: str):
        self._cellName = value

    @property
    def viewName(self):
        return self._viewName

    @viewName.setter
    def viewName(self, value: str):
        self._viewName = value

    @property
    def instanceName(self):
        return self._instanceName

    @instanceName.setter
    def instanceName(self, value: str):
        assert isinstance(value, str)
        self._instanceName = value

    @property
    def shapes(self):
        return self._shapes

    @shapes.setter
    def shapes(self, value: list[layoutShape]):
        self.removeShapes()
        self._shapes = value
        self.setShapes()

    @property
    def start(self):
        if self._start is None:
            self._start = self.childrenBoundingRect().bottomLeft()
        return self._start.toPoint()

    @property
    def counter(self):
        return self._counter

    @counter.setter
    def counter(self, value: int):
        if isinstance(value, int):
            self._counter = value

    def addShape(self, shape: layoutShape):
        self._shapes.append(shape)
        shape.setParentItem(self)


class layoutPcell(layoutInstance):
    def __init__(self, shapes: list):
        super().__init__(shapes)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._libraryName}, {self._cellName}, {self._viewName}, {self._instanceName})"


class layoutLine(layoutShape):
    def __init__(
            self,
            draftLine: QLineF,
            width: float = 1,
            mode: int = 0,
    ):
        """
        Initialize the TickLine object.

        Args:
            draftLine (QLineF): The draft line.
            width (float): The width of the line.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()

        self._draftLine = draftLine
        self._width: int = width
        self._mode = mode
        self._angle = 0
        self._rect = QRect(0, 0, 0, 0)
        penColour = QColor(255, 40, 40)
        self._pen = QPen(penColour, self._width, Qt.SolidLine)
        self._pen.setCosmetic(True)
        self._selectedPen = QPen(Qt.red, self._width + 1, Qt.SolidLine)
        self._selectedPen.setCosmetic(False)
        self._determineAngle(self._draftLine.angle())
        self.setZValue(999)

    def __repr__(self):
        return f"layoutLine({self._draftLine}, {self._width}, {self._mode})"

    def _determineAngle(self, angle: float):
        match self._mode:
            case 0:  # manhattan
                self._createManhattanLine(angle)
            case 1:  # diagonal
                self._createDiagonalLine(angle)
            case 2:
                self._createAnyAngleLine(angle)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createManhattanLine(self, angle):
        if 0 <= angle <= 45 or 360 > angle > 315:
            self._angle = 0
        elif 45 < angle <= 135:
            self._angle = 90
        elif 135 < angle <= 225:
            self._angle = 180
        elif 225 < angle <= 315:
            self._angle = 270

    def _createDiagonalLine(self, angle):
        if 0 <= angle <= 22.5 or 360 > angle > 337.5:
            self._angle = 0
        elif 22.5 < angle <= 67.5:
            self._angle = 45
        elif 67.5 < angle <= 112.5:
            self._angle = 90
        elif 112.5 < angle <= 157.5:
            self._angle = 135
        elif 157.5 < angle <= 202.5:
            self._angle = 180
        elif 202.5 < angle <= 247.5:
            self._angle = 225
        elif 247.5 < angle <= 292.5:
            self._angle = 270
        elif 292.5 < angle <= 337.5:
            self._angle = 315

    def _createAnyAngleLine(self, angle):
        self._angle = angle

    def boundingRect(self) -> QRectF:
        half_w = self._width // 2 + 2
        return (
            QRectF(self._draftLine.p1(), self._draftLine.p2())
            .normalized()
            .adjusted(-half_w, -half_w, half_w, half_w)
        )

    def paint(self, painter, option, widget=None):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.boundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._draftLine)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._determineAngle(angle)

    @property
    def angle(self) -> float:
        return self._angle

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self._mode = value

    @property
    def sceneEndPoints(self):
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]


# "layer", "name", "mode", "width", "startExtend", "endExtend"


class layoutPath(layoutShape):
    def __init__(
            self,
            draftLine: QLineF,
            layer: ddef.layLayer,
            width: float = 1.0,
            startExtend: int = 0,
            endExtend: int = 0,
            mode: int = 0,
    ):
        """
        Initialize the class instance.

        Args:
            draftLine (QLineF): The draft line.
            layer (ddef.layLayer): The layer.
            width (float, optional): The width. Defaults to 1.0.
            startExtend (int, optional): The start extend. Defaults to 0.
            endExtend (int, optional): The end extend. Defaults to 0.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()
        self.start = None
        self._draftLine = draftLine
        self._startExtend = startExtend
        self._endExtend = endExtend
        self._width = width
        self._layer = layer
        self._mode = mode
        self._name = ""
        self._stretch = False
        self._stretchSide = None
        self._definePensBrushes(self._layer)
        self._rect = QRectF(0, 0, 0, 0)
        self._angle = 0
        self._rectCorners(self._draftLine.angle())
        self.setZValue(self._layer.z)

    def __repr__(self):
        return (
            f"layoutPath({self._draftLine}, {self._layer}"
            f"{self._width}, {self._startExtend}, {self._endExtend}, {self._mode})"
        )

    def _rectCorners(self, angle: float):
        match self._mode:
            case 0:  # manhattan
                self._createManhattanPath(angle)
            case 1:  # diagonal
                self._createDiagonalPath(angle)
            case 2:
                self._createAnyAnglePath(angle)
            case 3:
                self._createHorizontalPath(angle)
            case 4:
                self._createVerticalPath(angle)
        self._draftLine.setAngle(0)
        self._rect = self._extractRect()
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createManhattanPath(self, angle: float) -> None:
        """
        Creates a Manhattan path based on the given angle.

        :param angle: The angle in degrees.
        :type angle: float

        :return: None
        """
        self._angle = 90 * math.floor(((angle + 45) % 360) / 90)

    def _createDiagonalPath(self, angle: float) -> None:
        """
        Creates a manhattan or diagonal path based on the given angle.
        Parameters:
            angle (float): The angle in degrees.
        Returns:
            None
        """
        self._angle = 45 * math.floor(((angle + 22.5) % 360) / 45)

    def _createAnyAnglePath(self, angle: float) -> None:
        self._angle = angle

    def _createHorizontalPath(self, angle: float) -> None:
        self._angle = 180 * math.floor(((angle + 90) % 360) / 180)

    def _createVerticalPath(self, angle: float) -> None:
        angle = angle % 360
        if 0 <= angle < 180:
            self._angle = 90
        else:
            self._angle = 270

    def _extractRect(self):
        direction = self._draftLine.p2() - self._draftLine.p1()
        if direction == QPoint(0, 0):  # when the mouse pressed first time
            rect = (
                QRectF(self._draftLine.p1(), self._draftLine.p2())
                .adjusted(-2, -2, 2, 2)
                .normalized()
            )
        else:
            direction /= direction.manhattanLength()
            perpendicular = QPointF(-direction.y(), direction.x())
            point1 = (
                    self._draftLine.p1()
                    + perpendicular * self._width * 0.5
                    - direction * self._startExtend
            ).toPoint()
            point2 = (
                    self._draftLine.p2()
                    - perpendicular * self._width * 0.5
                    + direction * self._endExtend
            ).toPoint()
            rect = QRectF(point1, point2).normalized()
        return rect

    def paint(self, painter, option, widget):
        # Get scale once and cache it
        scale = self.scene().views()[0].transform().m11()
        if self.isSelected():
            if self._stretch:
                painter.setPen(self._stretchPen)
                self._updateTransformedBrush(self._stretchBrush, scale)
            else:
                painter.setPen(self._selectedPen)
                self._updateTransformedBrush(self._selectedBrush, scale)
        else:
            painter.setPen(self._pen)
            self._updateTransformedBrush(self._brush, scale)
        painter.setBrush(self._transformedBrush)
        painter.drawLine(self._draftLine)
        painter.drawRect(self._rect)

    def boundingRect(self) -> QRectF:
        return self._rect.adjusted(-2, -2, 2, 2)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._rectCorners(angle)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width
        self.prepareGeometryChange()
        self._rectCorners(self._angle)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self.prepareGeometryChange()
        self._mode = value
        self._rectCorners(self._angle)

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def startExtend(self) -> int:
        return self._startExtend

    @startExtend.setter
    def startExtend(self, value: str):
        self.prepareGeometryChange()
        self._startExtend = value
        self._rect = self._extractRect()

    @property
    def endExtend(self) -> int:
        return self._endExtend

    @endExtend.setter
    def endExtend(self, value: str):
        self.prepareGeometryChange()
        self._endExtend = value
        self._rect = self._extractRect()

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value
        self.prepareGeometryChange()
        self._rect = self._extractRect()
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, layer: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = layer

    @property
    def sceneEndPoints(self):
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            eventPos = event.pos().toPoint()
            if self._stretch:
                if (
                        eventPos - self._draftLine.p1().toPoint()
                ).manhattanLength() <= self.scene().snapDistance:
                    self._stretchSide = "p1"
                    self.setCursor(Qt.SizeHorCursor)
                elif (
                        eventPos - self._draftLine.p2().toPoint()
                ).manhattanLength() <= self.scene().snapDistance:
                    self._stretchSide = "p2"
                    self.setCursor(Qt.SizeHorCursor)
                self.scene().stretchPath(self, self._stretchSide)


class layoutRuler(layoutShape):
    def __init__(
            self,
            draftLine: QLineF,
            width: float,
            tickGap: float,
            tickLength: int,
            tickFont: QFont,
            mode: int = 0,
    ):
        """
        Initialize the TickLine object.

        Args:
            draftLine (QLineF): The draft line.
            width (float): The width of the line.
            tickGap (float): The gap between ticks.
            tickLength (int): The length of the ticks.
            tickFont (QFont): The font for tick labels.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()

        self._draftLine = draftLine
        self._width = width
        self._tickGap = tickGap
        self._tickLength = tickLength
        self._mode = mode
        self._angle = 0
        self._rect = QRect(0, 0, 0, 0)
        penColour = QColor(255, 255, 40)
        self._pen = QPen(penColour, self._width, Qt.SolidLine)
        self._pen.setCosmetic(True)
        self._selectedPen = QPen(Qt.red, self._width + 1, Qt.SolidLine)
        self._selectedPen.setCosmetic(True)
        # self._pen.setCosmetic(True)
        self._tickTuples = list()
        self._tickFont = tickFont
        self._determineAngle(self._draftLine.angle())
        self._fm = QFontMetrics(self._tickFont)
        self._fontHeight = self._fm.boundingRect("0").height()
        self._fontWidth = self._fm.boundingRect("0.000").width()
        # Enable child event filtering for filters and handles
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        # Enable flag to indicate that the item contains children in shape
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        self._createRulerTicks()
        # self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
        # self.update(self.boundingRect())
        self.setZValue(999)

    def __repr__(self):
        return (
            f"layoutRuler({self._draftLine}, {self._width}, {self._tickGap}, "
            f"{self._tickLength}, {self._tickFont}, {self._mode})"
        )

    def _determineAngle(self, angle: float):
        match self._mode:
            case 0:  # manhattan
                self._createManhattanRuler(angle)
            case 1:  # diagonal
                self._createDiagonalRuler(angle)
            case 2:
                self._createAnyAngleRuler(angle)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createManhattanRuler(self, angle):
        if 0 <= angle <= 45 or 360 > angle > 315:
            self._angle = 0
        elif 45 < angle <= 135:
            self._angle = 90
        elif 135 < angle <= 225:
            self._angle = 180
        elif 225 < angle <= 315:
            self._angle = 270

    def _createDiagonalRuler(self, angle):
        if 0 <= angle <= 22.5 or 360 > angle > 337.5:
            self._angle = 0
        elif 22.5 < angle <= 67.5:
            self._angle = 45
        elif 67.5 < angle <= 112.5:
            self._angle = 90
        elif 112.5 < angle <= 157.5:
            self._angle = 135
        elif 157.5 < angle <= 202.5:
            self._angle = 180
        elif 202.5 < angle <= 247.5:
            self._angle = 225
        elif 247.5 < angle <= 292.5:
            self._angle = 270
        elif 292.5 < angle <= 337.5:
            self._angle = 315

    def _createAnyAngleRuler(self, angle):
        self._angle = angle

    def _createRulerTicks(self):
        self._tickTuples = []
        direction = self._draftLine.p2() - self._draftLine.p1()

        if direction != QPointF(0, 0):
            length = direction.manhattanLength()
            direction = QPointF(direction.x() / length, direction.y() / length)
            perpendicular = QPointF(-direction.y(), direction.x())

            if self._draftLine.length() >= self._tickGap:
                # Pre-calculate common values
                p1 = self._draftLine.p1()
                perp_tick = perpendicular * self._tickLength
                numberOfTicks = math.ceil(self._draftLine.length() / self._tickGap)

                # Generate ticks in single loop
                for i in range(numberOfTicks):
                    tick_pos = p1 + direction * (i * self._tickGap)
                    tick_line = QLineF(tick_pos, tick_pos + perp_tick)
                    self._tickTuples.append(
                        ddef.rulerTuple(
                            tick_pos + perp_tick,
                            (tick_line.p1(), tick_line.p2()),
                            str(float(i)),
                        )
                    )

            # Final tick
            p2 = self._draftLine.p2()
            final_line = QLineF(p2, p2 + perpendicular * self._tickLength)
            self._tickTuples.append(
                ddef.rulerTuple(
                    p2 + direction * 2,
                    (final_line.p1(), final_line.p2()),
                    str(round(self._draftLine.length() / self._tickGap, 3)),
                )
            )

        self._rect = QRectF(
            self._draftLine.p1().toPoint(), self._draftLine.p2().toPoint()
        ).normalized()

    def boundingRect(self) -> QRectF:
        margin = max(self._fontWidth, self._fontHeight, self._tickLength)
        return (
            QRectF(self._rect).normalized().adjusted(-margin, -margin, margin, margin)
        )

    def paint(self, painter, option, widget=None):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.childrenBoundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._draftLine)
        painter.setFont(self._tickFont)
        for tickTuple in self._tickTuples:
            painter.drawLine(tickTuple.line[0], tickTuple.line[1])
            painter.save()
            painter.translate(tickTuple.point)
            painter.rotate(self.angle)
            painter.translate(-tickTuple.point)
            painter.drawText(tickTuple.point.x(), tickTuple.point.y(), tickTuple.text)
            painter.restore()

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._determineAngle(angle)
        self._createRulerTicks()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self._mode = value

    @property
    def tickFont(self):
        return self._tickFont

    @property
    def tickGap(self):
        return self._tickGap


class layoutLabel(layoutShape):
    LABEL_ALIGNMENTS = ["Left", "Center", "Right"]
    LABEL_ORIENTS = ["R0", "R90", "R180", "R270", "MX", "MX90", "MY", "MY90"]
    LABEL_SCALE = 10

    def __init__(
            self,
            start: QPoint,
            labelText: str,
            fontFamily: str,
            fontStyle: str,
            fontHeight: str,
            labelAlign: str,
            labelOrient: str,
            layer: ddef.layLayer,
    ):
        super().__init__()
        self._start = start
        self._labelText = labelText
        self._fontFamily = fontFamily
        self._fontStyle = fontStyle
        self._fontHeight = fontHeight
        self._labelAlign = labelAlign
        self._labelOrient = labelOrient
        self._layer = layer
        self._definePensBrushes(self._layer)
        self._labelFont = QFont(fontFamily)
        self._labelFont.setStyleName(fontStyle)
        self._labelFont.setKerning(False)
        self._labelFont.setPointSize(int(float(self._fontHeight) * self.LABEL_SCALE))
        # self.setOpacity(1)
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelText)
        self._labelOptions = QTextOption()
        if self._labelAlign == layoutLabel.LABEL_ALIGNMENTS[0]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignLeft)
        elif self._labelAlign == layoutLabel.LABEL_ALIGNMENTS[1]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignCenter)
        elif self._labelAlign == layoutLabel.LABEL_ALIGNMENTS[2]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.setOrient()
        self.setZValue(self._layer.z)

    def __repr__(self):
        return (
            f"layoutLabel({self._start}, {self._labelText}, {self._fontFamily}, "
            f"{self._fontStyle}, {self._fontHeight}, {self._labelAlign}, "
            f"{self._labelOrient}, {self._layer})"
        )

    def setOrient(self):
        self.setTransformOriginPoint(self.mapFromScene(self._start))
        if self._labelOrient == layoutLabel.LABEL_ORIENTS[0]:
            self.setRotation(0)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[1]:
            self.setRotation(90)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[2]:
            self.setRotation(180)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[3]:
            self.setRotation(270)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[4]:
            self.flipTuple = (-1, 1)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[5]:
            self.flipTuple = (-1, 1)
            self.setRotation(90)
        elif self._labelOrient == layoutLabel.LABEL_ORIENTS[6]:
            self.flipTuple = (1, -1)
            self.setRotation(90)

    def boundingRect(self):
        return (
            QRect(
                self._start.x(),
                self._start.y(),
                self._rect.width(),
                self._rect.height(),
            )
            .normalized()
            .adjusted(-2, -2, 2, 2)
        )  #

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path

    def paint(self, painter, option, widget):
        painter.setFont(self._labelFont)
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.boundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawText(
            QPoint(self._start.x(), self._start.y() + self._rect.height()),
            self._labelText,
        )
        painter.drawPoint(self._start)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: QPoint):
        self.prepareGeometryChange()
        self._start = value

    @property
    def labelText(self):
        return self._labelText

    @labelText.setter
    def labelText(self, value):
        self.prepareGeometryChange()
        self._labelText = value
        self._rect = self._fm.boundingRect(self._labelText)

    @property
    def labelFont(self):
        return self._labelFont

    @property
    def fontFamily(self) -> str:
        return self._labelFont.family()

    @fontFamily.setter
    def fontFamily(self, familyName):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        if familyName in fixedFamilies:
            self._labelFont.setFamily(familyName)
        else:
            self.scene().logger.error(f"Not a valid font name: {familyName}")

    @property
    def fontStyle(self):
        return self._labelFont.styleName()

    @fontStyle.setter
    def fontStyle(self, value):
        if value in QFontDatabase.styles(self._labelFont.family()):
            self._labelFont.setStyleName(value)
        else:
            self.scene().logger.error(f"Not a valid font style: {value}")

    @property
    def fontHeight(self):
        return self._fontHeight

    @fontHeight.setter
    def fontHeight(self, value: str):
        self.prepareGeometryChange()
        self._fontHeight = value
        self._labelFont.setPointSize(int(float(self._fontHeight) * self.LABEL_SCALE))
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelText)

    @property
    def labelAlign(self):
        return self._labelAlign

    @labelAlign.setter
    def labelAlign(self, value):
        self.prepareGeometryChange()
        self._labelAlign = value

    @property
    def labelOrient(self):
        return self._labelOrient

    @labelOrient.setter
    def labelOrient(self, value):
        self.prepareGeometryChange()
        self._labelOrient = value


class layoutPin(layoutShape):
    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(
            self,
            start,
            end,
            pinName: str,
            pinDir: str,
            pinType: str,
            layer: ddef.layLayer,
    ):
        super().__init__()
        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._connected = False  # True if the pin is connected to a net.
        self._rect = QRect(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._layer = layer
        self._definePensBrushes(self._layer)
        self._label = None
        self._stretchSide = None
        self._stretchPen = QPen(QColor("red"), self._layer.pwidth, Qt.SolidLine)
        self.setZValue(self._layer.z)

    def __repr__(self):
        return (
            f"layoutPin({self._start}, {self._end}, {self._pinName}, {self._pinDir}, "
            f"{self._pinType}, {self._layer})"
        )

    def paint(self, painter, option, widget):
        # Get scale once and cache it
        scale = self.scene().views()[0].transform().m11()
        if self.isSelected():
            painter.setPen(self._selectedPen)
            self._updateTransformedBrush(self._selectedBrush, scale)
        else:
            painter.setPen(self._pen)
            self._updateTransformedBrush(self._brush, scale)
        painter.setBrush(self._brush)
        painter.drawRect(self._rect)

    def boundingRect(self):
        return self._rect.adjusted(-2, 2, 2, 2)

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, value):
        self._pinName = value

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, value):
        self._pinDir = value

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, value):
        self._pinType = value

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self._end).normalized()
        self._start = self._rect.topLeft()

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self._start, end).normalized()
        self._end = self._rect.bottomRight()

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value: layoutLabel):
        if isinstance(value, layoutLabel):
            self._label = value
        else:
            self.scene().logger.error("Not a Label")

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                if eventPos.x() == self._rect.left():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[0]
                elif eventPos.x() == self._rect.right():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[1]
                elif eventPos.y() == self._rect.top():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[2]
                elif eventPos.y() == self._rect.bottom():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[3]
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretch and self._stretchSide:
            self.prepareGeometryChange()
            if self.stretchSide == layoutRect.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == layoutRect.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self._pen.width() / 2))
            elif self.stretchSide == layoutRect.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == layoutRect.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self._pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class layoutVia(layoutShape):
    def __init__(
            self,
            start: QPoint,
            viaDefTuple: ddef.viaDefTuple,
            width: int,
            height: int,
    ):
        super().__init__()
        end = start + QPoint(width, height)
        self._rect = QRectF(start, end).normalized().toRect()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._viaDefTuple = viaDefTuple
        self._layer = viaDefTuple.layer
        self._type = viaDefTuple.type
        self._width = width
        self._height = height
        self._definePensBrushes(self._layer)
        self.setZValue(self._layer.z)

    def __repr__(self):
        return f"layoutVia({self._start}, {self._end}, {self._layer})"

    def paint(self, painter, option, widget):
        scale = self.scene().views()[0].transform().m11()
        if self.isSelected():
            painter.setPen(self._selectedPen)
        else:
            painter.setPen(self._pen)
        self._updateTransformedBrush(self._brush, scale)
        painter.setBrush(self._transformedBrush)
        painter.drawRect(self._rect)
        painter.drawLine(self._rect.bottomLeft(), self._rect.topRight())
        painter.drawLine(self._rect.topLeft(), self._rect.bottomRight())

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect.moveTo(self.mapFromScene(start))

    @property
    def width(self):
        return self._rect.width()

    @width.setter
    def width(self, value: int):
        self.prepareGeometryChange()
        self._rect.setWidth(value)

    @property
    def height(self):
        return self._rect.height()

    @height.setter
    def height(self, value: int):
        self._rect.setHeight(value)

    @property
    def viaDefTuple(self):
        return self._viaDefTuple

    @property
    def type(self):
        return self._type


class layoutViaArray(layoutShape):
    def __init__(
            self,
            start: QPoint,
            prototype_via,
            xs: float,
            ys: float,
            xnum: int,
            ynum: int,
    ):
        super().__init__()
        self._prototype_via = prototype_via
        self._ynum = ynum  # number of rows
        self._xnum = xnum  # number of columns
        self._xs = xs  # column spacing
        self._ys = ys  # row spacing
        self._start = start  # top-left corner location
        self._via = layoutVia(
            self._start,
            self._prototype_via.viaDefTuple,
            self._prototype_via.width,
            self._prototype_via.height,
        )
        self._via_array = []
        self._create_array()
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        self._selectedPen = QPen(QColor("yellow"), 4, Qt.DashLine)
        self._selectedPen.setCosmetic(True)

    def _create_array(self):
        # Pre-calculate constants
        x_step = self._xs + self._prototype_via.width
        y_step = self._ys + self._prototype_via.height
        start_x, start_y = self._start.x(), self._start.y()
        via_def = self._prototype_via.viaDefTuple
        via_width = self._prototype_via.width
        via_height = self._prototype_via.height

        # Create flat list using itertools.product
        vias = [
            self._create_via(
                start_x + col * x_step,
                start_y + row * y_step,
                via_def,
                via_width,
                via_height,
            )
            for row, col in itertools.product(range(self._ynum), range(self._xnum))
        ]

        # Reshape into 2D array
        self._via_array = [
            vias[i: i + self._xnum] for i in range(0, len(vias), self._xnum)
        ]

    def _create_via(self, x, y, via_def, width, height):
        via = layoutVia(QPoint(x, y), via_def, width, height)
        via.setFlag(QGraphicsItem.ItemIsSelectable, False)
        via.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
        via.setParentItem(self)
        return via

    def boundingRect(self) -> QRectF:
        return self.childrenBoundingRect()

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.childrenBoundingRect())
        return path

    def paint(self, painter, option, widget):
        painter.setRenderHint(QPainter.NonCosmeticBrushPatterns)
        if option.state & QStyle.State_Selected:
            painter.setPen(self._selectedPen)
            painter.drawRect(self.childrenBoundingRect())

    @property
    def via_array(self):
        return self._via_array

    @property
    def ynum(self):
        return self._ynum

    @property
    def xnum(self):
        return self._xnum

    @property
    def start(self) -> QPoint:
        return self._start

    @property
    def xs(self) -> float:
        return self._xs

    @property
    def ys(self):
        return self._ys

    @property
    def via(self):
        return self._via

    @property
    def height(self):
        return self._via.height

    @property
    def width(self):
        return self._via.width


class layoutPolygon(layoutShape):
    __slots__ = (
        "_points",
        "_layer",
        "_polygon",
        "_selectedCorner",
        "_selectedCornerIndex",
    )

    def __init__(self, points: list, layer: ddef.layLayer):
        super().__init__()
        self._points = points
        self._layer = layer
        self._selectedCorner = QPoint(99999, 99999)
        self._selectedCornerIndex = 999

        # Defer expensive operations
        self._polygon = None
        self._definePensBrushes(self._layer)
        self.setZValue(self._layer.z)
        self._flipTuple = (1, 1)  # Direct assignment instead of property

    def __repr__(self):
        return f"layoutPolygon({self._points}, {self._layer})"

    def paint(self, painter, option, widget):
        # Cache frequently accessed values
        selected = self.isSelected()
        scale = self.scene().views()[0].transform().m11()

        # Set pen and brush based on selection state
        if selected:
            painter.setPen(self._selectedPen)
            self._updateTransformedBrush(self._selectedBrush, scale)
            # Use cached corner coordinates (avoid QPoint creation)
            if self._stretch and self._selectedCorner.x() != 99999:
                painter.drawEllipse(self._selectedCorner, 5, 5)
        else:
            painter.setPen(self._pen)
            self._updateTransformedBrush(self._brush, scale)

        painter.setBrush(self._transformedBrush)
        painter.drawPolygon(self.polygon)

    def boundingRect(self) -> QRectF:
        return self.polygon.boundingRect()

    @property
    def polygon(self):
        if self._polygon is None:
            self._polygon = QPolygonF(self._points)
        return self._polygon

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, layer: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = layer

    @property
    def points(self) -> list:
        return self._points

    @points.setter
    def points(self, value: list):
        self.prepareGeometryChange()
        self._points = value
        self._polygon = None  # Invalidate cache

    def addPoint(self, point: QPoint):
        self.prepareGeometryChange()
        self._points.append(point)
        self._polygon = None  # Invalidate cache

    @property
    def tempLastPoint(self):
        return self._points[-1]

    @tempLastPoint.setter
    def tempLastPoint(self, value: QPoint):
        self.prepareGeometryChange()
        self._polygon = QPolygonF([*self._points, value])

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                for point in self._points:
                    if (
                            eventPos - point
                    ).manhattanLength() <= self.scene().snapDistance:
                        self._selectedCorner = point
                        self._selectedCornerIndex = self._points.index(point)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretch and self._selectedCornerIndex != 999:
            self._points[self._selectedCornerIndex] = eventPos
            self.points = self._points
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)
            self._selectedCornerIndex = 999


class alignLine(layoutShape):
    def __init__(
            self,
            draftLine: QLineF,
            width: int = 1,
            mode: int = 0,
    ):
        """
        Initialize the TickLine object.

        Args:
            draftLine (QLineF): The draft line.
            width (float): The width of the line.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()

        self._draftLine = draftLine
        self._width: int = width
        self._mode = mode
        self._angle = 0
        self._rect = QRect(0, 0, 0, 0)
        penColour = QColor(255, 40, 40)
        self._pen = QPen(penColour, self._width, Qt.SolidLine)
        self._pen.setCosmetic(True)
        self._selectedPen = QPen(Qt.red, self._width + 1, Qt.SolidLine)
        self._selectedPen.setCosmetic(True)
        self._determineAngle(self._draftLine.angle())
        self.setZValue(999)

    def __repr__(self):
        return f"alignLine({self._draftLine}, {self._width}, {self._mode})"

    def _determineAngle(self, angle: float):
        match self._mode:
            case 0:  # horizontal
                self._createHorizontalLine(angle)
            case 1:  # vertical
                self._createVerticalLine(angle)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createHorizontalLine(self, angle):
        if 0 <= angle <= 90 or 360 > angle > 270:
            self._angle = 0
        elif 90 < angle <= 270:
            self._angle = 180

    def _createVerticalLine(self, angle):
        if 0 <= angle < 180:
            self._angle = 90
        elif 180 < angle <= 360:
            self._angle = 270

    def boundingRect(self) -> QRectF:
        half_w = self._width // 2 + 2
        return (
            QRectF(self._draftLine.p1(), self._draftLine.p2())
            .normalized()
            .adjusted(-half_w, -half_w, half_w, half_w)
        )

    def paint(self, painter, option, widget=None):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.boundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._draftLine)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._determineAngle(angle)

    @property
    def angle(self) -> float:
        return self._angle

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self._mode = value

    @property
    def sceneEndPoints(self):
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json
from typing import Dict, Any

from PySide6.QtCore import QPointF

import revedaEditor.common.labels as lbl
import revedaEditor.common.shapes as shp


class symbolAttribute(object):
    def __init__(self, name: str, definition: str):
        self._name = name
        self._definition = definition

    def __str__(self):
        return f"{self.name}: {self.definition}"

    def __repr__(self):
        return f"{type(self)}({self.name},{self.definition})"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        assert isinstance(value, str)
        self._name = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        assert isinstance(value, str)
        self._definition = value


class symbolEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Dict[str, Any]:
        if isinstance(item, shp.symbolRectangle):
            return self._encodeSymbolRectangle(item)
        elif isinstance(item, shp.symbolLine):
            return self._encodeSymbolLine(item)
        elif isinstance(item, shp.symbolCircle):
            return self._encodeSymbolCircle(item)
        elif isinstance(item, shp.symbolArc):
            return self._encodeSymbolArc(item)
        elif isinstance(item, shp.symbolPolygon):
            return self._encodeSymbolPolygon(item)
        elif isinstance(item, shp.symbolPin):
            return self._encodeSymbolPin(item)
        elif isinstance(item, shp.text):
            return self._encodeText(item)
        elif isinstance(item, lbl.symbolLabel):
            return self._encodeSymbolLabel(item)
        elif isinstance(item, symbolAttribute):
            return self._encodeSymbolAttribute(item)
        return super().default(item)

    def _encodeSymbolRectangle(self, item: shp.symbolRectangle) -> Dict[str, Any]:
        return {
            "type": "rect",
            "rect": item.rect.getCoords(),
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeSymbolLine(self, item: shp.symbolLine) -> Dict[str, Any]:
        return {
            "type": "line",
            "st": item.start.toTuple(),
            "end": item.end.toTuple(),
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeSymbolCircle(self, item: shp.symbolCircle) -> Dict[str, Any]:
        return {
            "type": "circle",
            "cen": item.centre.toTuple(),
            "end": item.end.toTuple(),
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeSymbolArc(self, item: shp.symbolArc) -> Dict[str, Any]:
        return {
            "type": "arc",
            "st": item.start.toTuple(),
            "end": item.end.toTuple(),
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
            "at": shp.symbolArc.arcTypes.index(item.arcType),
        }

    def _encodeSymbolPolygon(self, item: shp.symbolPolygon) -> Dict[str, Any]:
        return {
            "type": "polygon",
            "ps": [item.mapToScene(p).toTuple() for p in item.points],
            "fl": item.flipTuple,
        }

    def _encodeSymbolPin(self, item: shp.symbolPin) -> Dict[str, Any]:
        return {
            "type": "pin",
            "st": item.start.toTuple(),
            "nam": item.pinName,
            "pd": item.pinDir,
            "pt": item.pinType,
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeText(self, item: shp.text) -> Dict[str, Any]:
        return {
            "type": "text",
            "st": item.start.toTuple(),
            "tc": item.textContent,
            "ff": item.fontFamily,
            "fs": item.fontStyle,
            "th": item.textHeight,
            "ta": item.textAlignment,
            "to": item.textOrient,
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeSymbolLabel(self, item: lbl.symbolLabel) -> Dict[str, Any]:
        return {
            "type": "label",
            "st": item.start.toTuple(),
            "nam": item.labelName,
            "def": item.labelDefinition,
            "txt": item.labelText,
            "val": item.labelValue,
            "vis": item.labelVisible,
            "lt": item.labelType,
            "ht": item.labelHeight,
            "al": item.labelAlign,
            "or": item.labelOrient,
            "use": item.labelUse,
            "loc": self._subtract_point(item.scenePos(), item.scene().origin),
            "fl": item.flipTuple,
        }

    def _encodeSymbolAttribute(self, item: symbolAttribute) -> Dict[str, Any]:
        return {
            "type": "attr",
            "nam": item.name,
            "def": item.definition,
        }

    @staticmethod
    def _subtract_point(point: QPointF, origin: QPointF) -> tuple:
        return (point - origin).toTuple()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   Add-ons and extensions developed for this software may be distributed
#
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)


# net class definition.
from enum import IntEnum
from functools import cached_property
from typing import Dict, List, Set, Union

from PySide6.QtCore import (
    QLineF,
    QPoint,
    QPointF,
    QRectF,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QPainter,
    QPainterPath,
    QPen,
)
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsLineItem,
    QGraphicsPathItem,
    QGraphicsSceneHoverEvent,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
)

from revedaEditor.backend.pdkLoader import importPDKModule

schlyr = importPDKModule("schLayers")


class NetMode(IntEnum):
    ORTHOGONAL = 0
    DIAGONAL = 1
    FREE = 2


class netNameStrengthEnum(IntEnum):
    NONAME = 0
    WEAK = 1
    INHERIT = 2
    SET = 3


class schematicNet(QGraphicsItem):
    def __init__(self, start: QPoint, end: QPoint, width: int = 0, mode: int = 0):
        super().__init__()

        # Batch set multiple flags at once using setFlags
        self.setFlags(
            QGraphicsItem.ItemIsSelectable
            | QGraphicsItem.ItemIsFocusable
            | QGraphicsItem.ItemSendsGeometryChanges
        )
        self.setFlag(QGraphicsItem.ItemIsMovable, False)

        self.setAcceptHoverEvents(True)

        # Basic properties
        self._mode = mode
        self._width = width
        self._flip = (1, 1)
        self._offset = QPoint(0, 0)

        # State flags
        # self._stretch: bool = False
        self._nameConflict: bool = False
        self._highlighted: bool = False
        self._nameStrength: int = netNameStrengthEnum.NONAME

        # Collections
        self._flightLinesSet: Set["schematicNet"] = set()
        self._connectedNetsSet: Set["schematicNet"] = set()
        self._endPointNetDict: Dict[int, Set["schematicNet"]]
        # track nets connected to each end point.
        self._netSnapLines: dict[QPoint, Set["schematicNet"]] = {}

        # Line and name initialization
        self.draftLine = QLineF(start, end)
        self._nameItem = self.createEmptyNameItem()

    def createEmptyNameItem(self):
        nameItem = netName("", self)
        nameItem.setPos(self.draftLine.center())
        nameItem.setParentItem(self)
        return nameItem

    @property
    def draftLine(self) -> QLineF:
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        # Snap endpoints to grid if scene is available
        scene = self.scene()
        if scene:
            p1 = scene.snapToGrid(line.p1().toPoint())
            p2 = scene.snapToGrid(line.p2().toPoint())
            line = QLineF(p1, p2)

        # Early exit if the line hasn't changed
        if hasattr(self, "_draftLine") and self._draftLine == line:
            return

        self.prepareGeometryChange()
        self.__dict__.pop("_hash_value", None)

        self._draftLine = line
        origin_point = line.p1()
        line_angle = line.angle()

        angle_calculators = {
            0: lambda angle: 90 * round(angle / 90),
            1: lambda angle: 45 * round(angle / 45),
        }
        self._angle = angle_calculators.get(self._mode, lambda angle: angle)(line_angle)

        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(origin_point)
        self.setRotation(-self._angle)

        # Clear the cached _extractRect
        self.__dict__.pop("_extractRect", None)
        self._shapeRect = self._extractRect.adjusted(-2, -2, 2, 2)
        self._boundingRect = self._extractRect.adjusted(-10, -10, 10, 10)

    @cached_property
    def _extractRect(self) -> QRectF:
        p1 = self._draftLine.p1()
        p2 = self._draftLine.p2()
        direction = p2 - p1

        # Early return for zero-length line
        if direction == QPoint(0, 0):
            return QRectF(p1, p2).adjusted(-2, -2, 2, 2).normalized()

        manhattan_length = direction.manhattanLength()
        if manhattan_length == 0:
            return QRectF(p1, p2).normalized()

        # Cache frequently used values and combine calculations
        half_width = 0.5 * self._getPen().width() * (self._width + 1)
        inv_length = 1.0 / manhattan_length

        # Calculate direction components in one step
        direction_x = direction.x() * inv_length
        direction_y = direction.y() * inv_length

        # Calculate perpendicular offsets
        perp_x = -direction_y * half_width
        perp_y = direction_x * half_width

        point1 = QPoint(
            round(
                p1.x() + perp_x
            ),  # Using round() instead of int() for better precision
            round(p1.y() + perp_y),
        )
        point2 = QPoint(round(p2.x() - perp_x), round(p2.y() - perp_y))

        return QRectF(point1, point2).normalized()

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self._shapeRect)
        return path

    def boundingRect(self):
        return self._boundingRect

    def paint(self, painter: QPainter, option, widget=None):
        pen = self._getPen()
        painter.setPen(pen)
        painter.drawLine(self._draftLine)

    def _getPen(self) -> QPen:
        pen_mapping = {
            self.isSelected(): schlyr.selectedWirePen,
            # self._stretch: schlyr.stretchWirePen,
            self._highlighted: schlyr.hilightPen,
            self._nameConflict: schlyr.errorWirePen,
        }
        base_pen = next(
            (pen for condition, pen in pen_mapping.items() if condition), schlyr.wirePen
        )
        return_pen = QPen(base_pen)
        return_pen.setWidth(base_pen.width() * (self._width + 1))
        return return_pen

    def __repr__(self):
        return f"schematicNet({self.sceneEndPoints}, {self._width})"

    def itemChange(self, change, value):
        if self.scene():
            match change:
                case QGraphicsItem.GraphicsItemChange.ItemSelectedHasChanged:
                    if value:
                        self.setZValue(self.zValue() + 10)
                        self.scene().selectedNet = self
                    else:
                        self.setZValue(self.zValue() - 10)
                        self.scene().selectedNet = None
                case QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged:
                    self.updateSnapLines()

        return super().itemChange(change, value)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        scene = self.scene()
        if scene:
            if scene.editModes.moveItem:
                self.setFlag(QGraphicsItem.ItemIsMovable, True)
            # elif self._stretch:
            #     eventPos = event.pos().toPoint()
            #     if (
            #             eventPos - self._draftLine.p1().toPoint()
            #     ).manhattanLength() <= scene.snapDistance:
            #         self.setCursor(Qt.SizeHorCursor)
            #         scene.stretchNet.emit(self, "p1")
            #     elif (
            #             eventPos - self._draftLine.p2().toPoint()
            #     ).manhattanLength() <= self.scene().snapDistance:
            #         self.setCursor(Qt.SizeHorCursor)
            #         scene.stretchNet.emit(self, "p2")
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setSelected(False)
        if self.scene().editModes.moveItem:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.scene().wireEditFinished.emit(self)
        super().mouseReleaseEvent(event)

    def generateEndPointNetDict(self):
        self._endPointNetDict = dict[int, Set[schematicNet]]()
        if self.scene():
            for i, endPoint in enumerate(self.sceneEndPoints):
                netSet: Set["schematicNet"] = {
                                                  item
                                                  for item in self.scene().items(endPoint)
                                                  if isinstance(item, schematicNet)
                                              } - {self}
                self._endPointNetDict[i] = netSet

    def initializeSnapLines(self):
        """Initialize snap lines for the net."""
        scene = self.scene()
        sceneGrid = scene.snapTuple[0]
        if not self._endPointNetDict:
            self.generateEndPointNetDict()
        for i, nets in self._endPointNetDict.items():
            point = self.sceneEndPoints[i]
            snapLinesSet: set["guideLine"] = set()
            for net in nets:
                for endPoint in net.sceneEndPoints:
                    if (endPoint - point).manhattanLength() < sceneGrid / 2:
                        otherEndPoint = net.sceneOtherEnd(endPoint)
                        snapLine = guideLine(point, otherEndPoint)
                        snapLine.inherit(net)
                        snapLinesSet.add(snapLine)
                        scene.deleteUndoStack(net)
                        scene.addItem(snapLine)
                        continue
            self._netSnapLines[i] = snapLinesSet

    def updateSnapLines(self):
        # Create a list to store lines for removal to avoid modifying the set during iteration
        lines_to_remove = []
        # shape is not connected to any nets
        for i, snapLinesSet in self._netSnapLines.items():
            point = self.sceneEndPoints[i]
            if not snapLinesSet:
                continue
            for snapLine in list(snapLinesSet):  # Iterate over a copy of the set
                current_end = snapLine.line().p2()
                newLine = QLineF(point, current_end)

                # Only update if there's a significant change
                if newLine.length() > 1 and (
                        abs(newLine.dx()) > 1 or abs(newLine.dy()) > 1
                ):
                    snapLine.setLine(newLine)
                else:
                    # Remove unnecessary lines
                    self.scene().removeItem(snapLine)
                    snapLinesSet.remove(snapLine)
                    lines_to_remove.append(snapLine)

        # Now, safely remove the collected items
        for snapLine in lines_to_remove:
            for snapLinesSet in self._netSnapLines.values():
                if snapLine in snapLinesSet:
                    snapLinesSet.remove(snapLine)

    def _finishSnapLines(self):
        if self._netSnapLines:
            try:
                scene = self.scene()
                for snapLinesSet in self._netSnapLines.values():
                    for snapLine in snapLinesSet:
                        newNets = scene.addStretchWires(
                            snapLine.line().p1().toPoint(),
                            snapLine.line().p2().toPoint(),
                        )
                        if newNets:
                            for netItem in newNets:
                                netItem.inherit(snapLine)
                            scene.addListUndoStack(newNets)
                        scene.removeItem(snapLine)
                self._netSnapLines = {}
            except Exception as e:
                # Log error but continue processing other guidelines
                scene.logger.error(f"Error processing snap lines: {e}")

    @staticmethod
    def parseArrayNotation(name: str) -> tuple[str, tuple[int, int]]:
        """Parse net/instance array notation like 'name<0:5>' into base name and index range.
        Also handles single instance/net notation like 'name<0>' or 'name<1>'.

        Returns:
            tuple[str, tuple[int, int]]: Base name and (start, end) indices.
                                         Returns (-1, -1) if no bus notation.
        """
        if "<" not in name or ">" not in name:
            return name, (-1, -1)

        baseName = name.split("<")[0]
        indexRange = name.split("<")[1].split(">")[0]

        if ":" not in indexRange:
            singleIndex = int(indexRange)
            return baseName, (singleIndex, singleIndex)

        start, end = map(int, indexRange.split(":"))
        return baseName, (start, end)

    @staticmethod
    def _namesMatch(name1: str, name2: str) -> bool:
        """Check if two net names match, considering bus notation.
        Examples:
        - 'inp' matches 'inp' (exact match)
        - 'inp<7>' matches 'inp<7:0>' (index in range)
        - 'data<3>' matches 'data<7:0>' (index in range)
        """
        if name1 == name2:
            return True

        base1, (start1, end1) = schematicNet.parseArrayNotation(name1)
        base2, (start2, end2) = schematicNet.parseArrayNotation(name2)

        # Base names must match
        if base1 != base2:
            return False

        # If both have no index (-1, -1), base names already matched above
        # If only one has index, they don't match (e.g., 'inp' vs 'inp<7>')
        if (start1 == -1) or (start2 == -1):
            return False

        # Check if index/range overlaps
        min1, max1 = min(start1, end1), max(start1, end1)
        min2, max2 = min(start2, end2), max(start2, end2)

        return not (max1 < min2 or max2 < min1)

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        """
        Override the hoverEnterEvent method of QGraphicsItem.

        Args:
            event (QGraphicsSceneHoverEvent): The hover event.

        Returns:
            None
        """
        super().hoverEnterEvent(event)
        # Check if highlightNets flag is set in the scene
        if self.scene().highlightNets:
            self._highlighted = True
            self.scene().nameSceneNets()  # first name all the nets in the scene.
            sceneNetsSet = self.scene().findSceneNetsSet()
            self._connectedNetsSet = {
                net for net in sceneNetsSet if self._namesMatch(net.name, self.name)
            }

            # Highlight the connected netItems
            for netItem in self._connectedNetsSet:
                netItem.highlight()
                flightLine = netFlightLine(
                    self.mapToScene(self._draftLine.center()),
                    netItem.mapToScene(netItem.draftLine.center()),
                )
                self._flightLinesSet.add(flightLine)
                self.scene().addItem(flightLine)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        if self._highlighted:
            self._highlighted = False
            self._clearFlightLines()
            self._unhighlightConnectedNets()

    def _clearFlightLines(self) -> None:
        for flight_line in self._flightLinesSet:
            self.scene().removeItem(flight_line)
        self._flightLinesSet.clear()

    def _unhighlightConnectedNets(self) -> None:
        for net_item in self._connectedNetsSet:
            net_item.unhighlight()

    def isParallel(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle) % 180) < 1

    def isOrthogonal(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle - 90) % 180) < 1

    def notParallel(self, otherNet: "schematicNet") -> bool:
        return not self.isParallel(otherNet)

    def findOverlapNets(self) -> Set["schematicNet"]:
        """
        Find all netItems in the scene that overlap with self.sceneShapeRect.

        Returns:
            set: A set of netItems that overlap with self.sceneShapeRect.
        """
        if self.scene():
            overlapNets = {
                netItem
                for netItem in self.collidingItems()
                if isinstance(netItem, schematicNet)
            }
            return overlapNets - {self}

    def inheritNetName(self, otherNet: "schematicNet") -> bool:
        """
        Inherit or resolve net names based on name strength and handle conflicts.
        """

        def resolve_name(weakNet: "schematicNet", strongNet: "schematicNet"):
            weakNet.name = strongNet.name
            weakNet.nameStrength = netNameStrengthEnum.INHERIT

        self_strength = self.nameStrength.value
        other_strength = otherNet.nameStrength.value

        if self_strength == 3:  # SET
            if other_strength < 3:
                resolve_name(otherNet, self)
                return True
            if self.name != otherNet.name:
                self.nameConflict = otherNet.nameConflict = True
                return False
            return True

        if self_strength == 2:  # INHERIT
            match other_strength:
                case 0 | 1:  # NONAME or WEAK
                    resolve_name(otherNet, self)
                    return True
                case 2:  # INHERIT
                    if self.name != otherNet.name:
                        self.nameConflict = otherNet.nameConflict = True
                        return False
                    return True
                case 3:
                    resolve_name(self, otherNet)
                    return True

        if self_strength == 1:  # WEAK
            match other_strength:
                case 0:  # NONAME
                    otherNet.nameStrength = netNameStrengthEnum.WEAK
                    otherNet.name = self.name
                    return True
                case 1:  # WEAK
                    if self.name != otherNet.name:
                        self.nameConflict = otherNet.nameConflict = True
                        return False
                    return True
                case 2 | 3:  # INHERIT or SET
                    resolve_name(self, otherNet)
                    return True

        if self_strength == 0:  # NONAME
            if other_strength > 0:
                resolve_name(self, otherNet)
                return True
            else:
                if self.name != "" or otherNet.name != "":
                    self.nameConflict = otherNet.nameConflict = True
                    return False
                return True

        return False

    def mergeNetName(self, otherNet: "schematicNet") -> bool:
        """
        Merge net names based on name strength and handle conflicts.
        """
        if otherNet.nameStrength > self.nameStrength:
            if otherNet.nameStrength == 3:
                self.name = otherNet.name
                self.nameStrength = netNameStrengthEnum.INHERIT
                return True
            else:
                self.name = otherNet.name
                self.nameStrength = otherNet.nameStrength
                return True
        elif otherNet.nameStrength == self.nameStrength:
            if otherNet.name != self.name:
                self.nameConflict = otherNet.nameConflict = True
                return False
            return True

    def clearName(self):
        """
        Clear the net name and set its strength to NONAME.
        """
        if self.nameStrength.value < 3:
            self.name = ""
            self.nameStrength = netNameStrengthEnum.NONAME

    def inherit(self, otherNet):
        self.name = otherNet.name
        self.nameStrength = otherNet.nameStrength

    @property
    def name(self) -> str:
        if self._nameItem:
            return self._nameItem.name
        else:
            return ""

    @name.setter
    def name(self, name: str):
        if name:
            self.prepareGeometryChange()
            self._nameItem.name = name
            self._nameItem.nameStrength = netNameStrengthEnum.SET
            self._nameItem.setPos(self._draftLine.center())

    @property
    def nameStrength(self) -> int:
        return self._nameItem.nameStrength

    @nameStrength.setter
    def nameStrength(self, value: netNameStrengthEnum):
        self._nameItem.nameStrength = value

    @property
    def nameConflict(self) -> bool:
        return self._nameConflict

    @nameConflict.setter
    def nameConflict(self, value: bool):
        self._nameConflict = value
        self._nameItem.nameConflict = value
        self.update()

    @property
    def endPoints(self):
        return [self._draftLine.p1().toPoint(), self._draftLine.p2().toPoint()]

    def otherEnd(self, point: QPoint) -> QPoint:
        if point == self._draftLine.p1().toPoint():
            return self._draftLine.p2().toPoint()
        else:
            return self._draftLine.p1().toPoint()

    def sceneOtherEnd(self, point: QPoint) -> QPoint:
        return self.mapToScene(self.otherEnd(point)).toPoint()

    @property
    def sceneEndPoints(self) -> List[QPoint]:
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]

    @sceneEndPoints.setter
    def sceneEndPoints(self, points: List[QPoint]):
        self.prepareGeometryChange()
        self.draftLine = QLineF(points[0], points[1])

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value: Union[QPoint | QPointF]):
        self._offset = value

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value: int):
        self.prepareGeometryChange()
        self._width = value
        # Invalidate cached hash when width changes
        if hasattr(self, "_hash_value"):
            del self._hash_value
        # Clear the cached _extractRect
        if "_extractRect" in self.__dict__:
            del self.__dict__["_extractRect"]
        self.update()

    def highlight(self):
        self._highlighted = True
        self.update()

    def unhighlight(self):
        self._highlighted = False
        self.update()

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, value: bool):
        self._highlighted = value

    @property
    def sceneShapeRect(self) -> QRectF:
        return self.mapRectToScene(self._shapeRect).normalized().toRect()

    # @property
    # def stretchSide(self) -> int:
    #     """
    #     The end where the net is stretched.
    #     """
    #     return self._stretchSide
    #
    # @stretchSide.setter
    # def stretchSide(self, value: str):
    #     self.prepareGeometryChange()
    #     self._stretchSide = value

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value

    # @property
    # def stretch(self) -> bool:
    #     return self._stretch
    #
    # @stretch.setter
    # def stretch(self, value: bool):
    #     self._stretch = value

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def nameItem(self) -> "netName":
        return self._nameItem

    @nameItem.setter
    def nameItem(self, name: str):
        self._nameItem = netName(name, self)
        self._nameItem.setPos(self._draftLine.center())
        self._nameItem.setParentItem(self)

    @property
    def annotationItem(self) -> "netAnnotation":
        return self._annotationItem

    @annotationItem.setter
    def annotationItem(self, text: str):
        self._annotationItem = netAnnotation(text, self)
        self._annotationItem.setPos(self._draftLine.center() + QPointF(0, -20))
        self._annotationItem.setParentItem(self)

    def clearAnnotation(self):
        if hasattr(self, "_annotationItem") and self._annotationItem:
            if self._annotationItem.scene():
                self._annotationItem.scene().removeItem(self._annotationItem)
            self._annotationItem = None

    def __hash__(self) -> int:
        """Generate a hash value for the schematic net based on its properties."""
        # Cache the hash value since the line length is unlikely to change frequently
        if not hasattr(self, "_hash_value"):
            # Combine multiple properties for a more unique hash
            self._hash_value = hash(
                (
                    self.draftLine.length(),
                    self.width,
                    frozenset((p.x(), p.y()) for p in self.sceneEndPoints),
                )
            )
        return self._hash_value

    def __eq__(self, other) -> bool:
        """Compare two schematic nets for equality."""
        if not isinstance(other, schematicNet):
            return False

        # Compare width first as it's a simple comparison
        if self.width != other.width:
            return False

        # Use set comprehension for endpoints comparison
        # Cache the results to avoid multiple set creations
        self_points = frozenset((point.x(), point.y()) for point in self.sceneEndPoints)
        other_points = frozenset(
            (point.x(), point.y()) for point in other.sceneEndPoints
        )

        return self_points == other_points


class netAnnotation(QGraphicsSimpleTextItem):
    def __init__(self, text: str, parent: schematicNet | None = None):
        super().__init__(text, parent)
        self._parent = parent
        self.setBrush(QBrush(QColor("orange")))
        if self._parent:
            self.setRotation(self._parent.angle)

    @property
    def parent(self):
        return self._parent

    @property
    def value(self) -> str:
        return self.text()

    @value.setter
    def value(self, input: str):
        self.prepareGeometryChange()
        self.setText(input)
        self.setRotation(self._parent.angle)
        self.parentItem().update()


class netName(QGraphicsSimpleTextItem):
    def __init__(self, name: str, parent: schematicNet | None = None):
        super().__init__(name, parent)
        self._parent = parent
        self.setBrush(schlyr.wireBrush)
        self._nameConflict = False
        self._nameStrength = netNameStrengthEnum.NONAME
        if self._parent:
            self.setRotation(self._parent.angle)
            self._draftLineCenter = self._parent.draftLine.center()

    def paint(self, painter, option, widget, /):
        if self._nameStrength == netNameStrengthEnum.SET:
            super().paint(painter, option, widget)
        else:
            # Do not paint the name if the strength is not SET
            pass

    @property
    def nameConflict(self) -> bool:
        return self._nameConflict

    @nameConflict.setter
    def nameConflict(self, value: bool):
        if value:
            self.setBrush(schlyr.errorWireBrush)
        else:
            self.setBrush(schlyr.wireBrush)
        self.update()

    @property
    def nameStrength(self) -> int:
        return self._nameStrength

    @nameStrength.setter
    def nameStrength(self, value: int):
        self._nameStrength = value
        if self._nameStrength == netNameStrengthEnum.SET:
            self.setVisible(True)
        else:
            self.setVisible(False)

    @property
    def name(self) -> str:
        return self.text()

    @name.setter
    def name(self, value: str):
        self.prepareGeometryChange()
        self.setText(value)
        self.setRotation(self._parent.angle)
        self.parentItem().update()

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent: schematicNet):
        self._parent = parent
        self.setRotation(self._parent.angle)
        self._draftLineCenter = self._parent.draftLine.center()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setSelected(False)
        self.setFlag(QGraphicsItem.ItemIsMovable, False)
        if self._parent:
            self._parent.unhighlight()
        super().mouseReleaseEvent(event)  # type: ignore


class netFlightLine(QGraphicsPathItem):
    wireHighlightPen = QPen(
        schlyr.wireHilightLayer.pcolor,
        schlyr.wireHilightLayer.pwidth,
        schlyr.wireHilightLayer.pstyle,
    )

    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__()
        self._createPath()

    def __repr__(self):
        return f"netFlightLine({self.mapToScene(self._start)},{self.mapToScene(self._end)})"

    def _createPath(self) -> None:
        line = QLineF(self._start, self._end)
        perpendicularLine = QLineF(
            line.center(), line.center() + QPointF(-line.dy(), line.dx())
        )
        perpendicularLine.setLength(100)

        path = QPainterPath()
        path.moveTo(self._start)
        path.quadTo(perpendicularLine.p2(), self._end)
        self.setPath(path)

    def paint(self, painter: QPainter, *_) -> None:
        painter.setPen(netFlightLine.wireHighlightPen)
        painter.drawPath(self.path())


class guideLine(QGraphicsLineItem):
    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__(QLineF(self._start, self._end))
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setPen(schlyr.guideLinePen)
        self._name: str = ""
        self._nameStrength: netNameStrengthEnum = netNameStrengthEnum.NONAME

    @property
    def sceneEndPoints(self) -> list[QPoint]:
        """
        Returns a list of the end points of the net in scene coordinates.

        Returns:
            list[QPoint]: A list of the end points of the net in scene coordinates.
        """
        return [
            self.mapToScene(self.line().p1()).toPoint(),
            self.mapToScene(self.line().p2()).toPoint(),
        ]

    def __repr__(self):
        return f"guideLine({self.mapToScene(self.line().p1()), self.mapToScene(self.line().p2())}"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "":  # net name should not be an empty string
            self._name = name

    @property
    def nameStrength(self) -> netNameStrengthEnum:
        return self._nameStrength

    @nameStrength.setter
    def nameStrength(self, value: netNameStrengthEnum):
        assert isinstance(value, netNameStrengthEnum)
        self._nameStrength = value

    def inherit(self, otherNet: schematicNet):
        """
        This method is used to carry the name information of the original net
        to stretch net.
        """
        assert isinstance(otherNet, schematicNet)
        self.name = otherNet.name
        self.nameStrength = otherNet.nameStrength

    def _finishSnapLines(self):
        if self._netSnapLines:
            scene = self.scene()
            for snapLinesSet in self._netSnapLines.values():
                for snapLine in snapLinesSet:
                    newNets = scene.addStretchWires(
                        snapLine.line().p1().toPoint(), snapLine.line().p2().toPoint()
                    )
                    if newNets:
                        for netItem in newNets:
                            netItem.inherit(snapLine)
                        scene.addListUndoStack(newNets)
                    scene.removeItem(snapLine)
            self._netSnapLines = dict()



import json
import pathlib

import gdstk
from PySide6.QtCore import (
    QPoint,
)
from PySide6.QtWidgets import (
    QMainWindow,
)
from numpy import pi

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.common.layoutShapes as lshp
import revedaEditor.fileio.layoutEncoder as layenc
from revedaEditor.backend.pdkLoader import importPDKModule

fabproc = importPDKModule("process")
laylyr = importPDKModule("layoutLayers")

dbu = float(fabproc.dbu)
snapGrid = float(fabproc.snapGrid)
majorGrid = float(fabproc.majorGrid)
gdsUnit = float(fabproc.gdsUnit)
gdsPrecision = float(fabproc.gdsPrecision)


class gdsImporter:
    def __init__(
            self,
            parent: QMainWindow,
            inputFile: pathlib.Path,
            importLibItem: libb.libraryItem,
    ):
        self._parent = parent
        self.inputFile = inputFile
        self._gdsLibrary = gdstk.read_gds(str(inputFile))
        self._gdsLibrary.set_property("name", str(inputFile.stem))
        self._libraryModel = self._parent.libraryBrowser.designView.libraryModel
        self._libItem = importLibItem

        self._topCells = self._gdsLibrary.top_level()
        self._unit = 1

    def importGDS(self):

        for cell in self._topCells:
            cellPath = self._libItem.libraryPath.joinpath(cell.name)
            cellItem = libb.createNewCellItem(self._libItem, cellPath)
            viewPath = cellItem.cellPath.joinpath("layout.json")
            viewItem = libb.createCellviewItem("layout", viewPath)
            self._processInstance(cell, viewItem)
        self._parent.logger.info(f"Imported {self.inputFile.stem} GDS File")
        self._parent.libraryBrowser.designView.reworkDesignLibrariesView(
            self._parent.libraryBrowser.designView.libraryModel.libraryDict)

    def _processInstance(self, cell: gdstk.Cell, viewItem: libb.viewItem):
        # Open file in context manager and write header
        with viewItem.viewPath.open("w") as file:
            file.write("[\n")
            file.write('    {"viewType": "layout"},\n')
            gridString = f"[{fabproc.majorGrid}, {fabproc.snapGrid}]"
            snapGridLine = '    {"snapGrid": ' + gridString + '},\n'
            file.write(snapGridLine)
            # Track if we need to write comma between items
            need_comma = False

            # Process instances
            for shape in self._processShapes(cell, viewItem):
                if need_comma:
                    file.write(",\n")
                json.dump(shape, file, cls=layenc.gdsImportEncoder, indent=4)
                need_comma = True

            # Close the JSON array
            file.write("\n]")

        return True  # Or return some status if needed

    def _processShapes(self, cell: gdstk.Cell, viewItem: libb.viewItem):
        """Generator that yields shapes one at a time."""
        # Process references
        for ref in cell.references:
            cellPath = self._libItem.libraryPath.joinpath(ref.cell_name)
            cellItem = libb.createNewCellItem(self._libItem, cellPath)
            viewPath = cellItem.cellPath.joinpath("layout.json")
            # Process nested instance first
            self._processInstance(ref.cell, libb.createCellviewItem("layout", viewPath))

            layoutInstance = lshp.layoutInstance([])
            layoutInstance.libraryName = cellItem.parent().libraryName
            layoutInstance.cellName = cellItem.cellName
            layoutInstance.viewName = viewItem.viewName
            layoutInstance.counter = 1
            layoutInstance.instanceName = 'I1'
            layoutInstance.setPos(ref.origin[0] * dbu, ref.origin[1] * dbu)
            layoutInstance.angle = ref.rotation * 180 / pi
            layoutInstance.flipTuple = (1, 1)
            yield layoutInstance

        # Process polygons
        for polygon in cell.polygons:
            layoutLayer = ddef.layLayer.filterByGDSLayer(
                laylyr.pdkAllLayers, polygon.layer, polygon.datatype
            )
            if layoutLayer:
                points = [QPoint(point[0] * dbu, point[1] * dbu) for point in
                          polygon.points]
                yield lshp.layoutPolygon(points, layoutLayer)

        # Process paths
        for path in cell.paths:
            for polygon in path.to_polygons():
                layoutLayer = ddef.layLayer.filterByGDSLayer(
                    laylyr.pdkAllLayers, polygon.layer, polygon.datatype
                )
                if layoutLayer:
                    points = [QPoint(point[0] * dbu, point[1] * dbu) for point in
                              polygon.points]
                    yield lshp.layoutPolygon(points, layoutLayer)
        # Process labels
        for gds_label in cell.labels:
            textLayer = ddef.layLayer.filterByGDSLayer(
                laylyr.pdkAllLayers, gds_label.layer, 0)
            if not textLayer:
                continue
            origin = QPoint(gds_label.origin[0] * dbu, gds_label.origin[1] * dbu)
            angle = gds_label.rotation * 180 / pi
            layout_label = lshp.layoutLabel(origin, gds_label.text, "Arial", "Regular",
                                            "10", "Center", "R0", textLayer)
            layout_label.angle = angle
            yield layout_label
